# Generated by Django 4.2.1 on 2024-02-08 07:55

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("simple_openid_connect_django", "0002_move_sessions_id_token"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidsession",
            name="raw_id_token",
            field=models.TextField(blank=True),
        ),
    ]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import sys
from pathlib import Path

import simple_openid_connect

BASE_PATH = Path(__file__).parent.parent
sys.path.append(str(BASE_PATH / "src"))


project = "simple_openid_connect"
copyright = "2022, Finn-Thorben Sell"
author = "Finn-Thorben Sell"
release = "v" + simple_openid_connect.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "django": (
        "https://docs.djangoproject.com/en/stable/",
        "https://docs.djangoproject.com/en/stable/_objects/",
    ),
    "drf-spectacular": ("https://drf-spectacular.readthedocs.io/en/latest/", None),
}
intersphinx_disabled_domains = ["std"]

autodoc_default_options = {
    "members": True,
    "special-members": "__init__",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



from simple_openid_connect.client_authentication import ClientSecretBasicAuth
from simple_openid_connect.flows import authorization_code_flow
from simple_openid_connect.flows.authorization_code_flow import (
    AuthenticationSuccessResponse,
)


def test_authorization_request(user_agent, dummy_auth_response):
    # act
    url = authorization_code_flow.start_authentication(
        "https://provider.example.com/auth",
        "openid",
        "test-client-id",
        "https://app.example.com/login-callback",
    )
    response = user_agent.naviagte_to(url)
    response_msg = authorization_code_flow.AuthenticationSuccessResponse.parse_url(
        response.url
    )

    # assert
    assert response_msg.code == "code.foobar123"


def test_token_exchange(user_agent, dummy_token_response):
    # act
    response = authorization_code_flow.exchange_code_for_tokens(
        "https://provider.example.com/token",
        AuthenticationSuccessResponse(code="code.foobar123"),
        "https://app.example.com/login-callback",
        ClientSecretBasicAuth(
            "test-client-id",
            "test-client-secret",
        ),
    )

    # assert
    assert response.access_token
    assert response.id_token


def test_handle_authentication_result(user_agent, dummy_token_response):
    # act
    response = authorization_code_flow.handle_authentication_result(
        "https://app.example.com/login-callback?code=code.foobar123",
        "https://provider.example.com/token",
        ClientSecretBasicAuth("test-client-id", "test-client-secret"),
    )

    # assert
    assert response.access_token == "access_token.foobar123"
    assert response.id_token == "id_token.user1"



"""
The `*Client Credentials Grant* <https://oauth.net/2/grant-types/client-credentials/>`_ (sometimes called Service Account Authentication) implementation.

This grant enables a client to retrieve tokens dedicated to the client and not to a specific user.
"""
import logging
from typing import Union

import requests

from simple_openid_connect.client_authentication import ClientAuthenticationMethod
from simple_openid_connect.data import (
    TokenErrorResponse,
    TokenRequest,
    TokenSuccessResponse,
)

logger = logging.getLogger(__name__)


def authenticate(
    token_endpoint: str, scope: str, client_authentication: ClientAuthenticationMethod
) -> Union[TokenSuccessResponse, TokenErrorResponse]:
    """
    Retrieve a token that is dedicated to the authenticated client from the provider.

    :param token_endpoint: The endpoint of the OP at which tokens can be exchanged.
        Corresponds to :data:`ProviderMetadata.token_endpoint <simple_openid_connect.data.ProviderMetadata.token_endpoint>`.
    :param scope: The scope requested by the application.
    :param client_authentication: A way for the client to authenticate itself.

    :returns: The result of the exchange
    """
    logger.debug(
        f"requesting access via client credentials grant as client {client_authentication.client_id}"
    )
    request_msg = TokenRequest(
        grant_type="client_credentials",
        scope=scope,
    )
    response = requests.post(
        token_endpoint,
        data=request_msg.encode_x_www_form_urlencoded(),
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        auth=client_authentication,
    )

    if response.status_code == 200:
        return TokenSuccessResponse.parse_raw(response.content)
    else:
        return TokenErrorResponse.parse_raw(response.content)



from typing import TYPE_CHECKING, List, Mapping, Optional, Union

from furl import furl

from simple_openid_connect.data import (
    AuthenticationSuccessResponse,
    TokenErrorResponse,
    TokenSuccessResponse,
)
from simple_openid_connect.exceptions import (
    ImpossibleOperationError,
    UnsupportedByProviderError,
)
from simple_openid_connect.flows import authorization_code_flow as impl

if TYPE_CHECKING:
    from simple_openid_connect.client import OpenidClient


class AuthorizationCodeFlowClient:
    """
    A client that implements *authorization code flow* related functionality.

    It is exposed via :data:`OpenidClient.authorization_code_flow <simple_openid_connect.client.OpenidClient.authorization_code_flow>`.
    """

    def __init__(self, base_client: "OpenidClient"):
        self._base_client = base_client

    def start_authentication(self) -> str:
        """
        Start the authentication process by constructing an appropriate :class:`AuthenticationRequest`, serializing it and
        returning a which the end user now needs to visit.

        :raises ImpossibleOperationError: If the client has no redirect_uri configured and therefore cannot perform this operation.

        :returns: A URL to which the user agent should be redirected
        """
        if self._base_client.authentication_redirect_uri is None:
            raise ImpossibleOperationError(
                "The client has no redirect_uri configured so no authentication flow can be started"
            )

        redirect_uri = furl(self._base_client.authentication_redirect_uri)

        return impl.start_authentication(
            self._base_client.provider_config.authorization_endpoint,
            self._base_client.scope,
            self._base_client.client_auth.client_id,
            redirect_uri.tostr(),
        )

    def handle_authentication_result(
        self,
        current_url: str,
        additional_redirect_args: Optional[Mapping[str, str]] = None,
    ) -> Union[TokenSuccessResponse, TokenErrorResponse]:
        """
        Handle an authentication result that is communicated to the RP in form of the user agents current url after having started an authentication process via :func:`start_authentication`.

        :param current_url: The current URL which the user is visiting.
            The authentication result should be encoded into this url by the authorization server.
        :param additional_redirect_args: Additional URL parameters that were added to the redirect uri.
            They are probably still present in `current_url` but since they could be of any shape, no attempt is made here to automatically reconstruct them.

        :raises AuthenticationFailedError: If the current url indicates an authentication failure that prevents an access token from being retrieved.
        :raises UnsupportedByProviderError: If the provider only supports implicit flow and has no token endpoint.
        :raises ImpossibleOperationError: If the client has no redirect_uri configured and therefore cannot perform this operation.

        :returns: The result of the token exchange
        """
        if self._base_client.provider_config.token_endpoint is None:
            raise UnsupportedByProviderError(
                f"The OpenID provider {self._base_client.provider_config.issuer} only supports the implicit flow and does not have a token endpoint"
            )

        if self._base_client.authentication_redirect_uri is None:
            raise ImpossibleOperationError(
                "The client has no redirect_uri configured so the authentication result cannot be handled correctly"
            )

        redirect_uri = furl(self._base_client.authentication_redirect_uri)
        if additional_redirect_args is not None:
            redirect_uri.args.update(additional_redirect_args)

        return impl.handle_authentication_result(
            current_url=current_url,
            token_endpoint=self._base_client.provider_config.token_endpoint,
            client_authentication=self._base_client.client_auth,
            redirect_uri=redirect_uri.tostr(),
        )

    def exchange_code_for_tokens(
        self, authentication_response: AuthenticationSuccessResponse
    ) -> Union[TokenSuccessResponse, TokenErrorResponse]:
        """
        Exchange a received code for access, refresh and id tokens.

        You might want to use :func:`handle_authentication_result` if you don't want to parse an authentication result from the users current url yourself.

        :param authentication_response: The (successful) response which this app received after the user has come back from the OP.

        :raises UnsupportedByProviderError: If the provider only supports implicit flow and has no token endpoint.
        :raises ImpossibleOperationError: If the client has no redirect_uri configured and therefore cannot perform this operation.

        :returns: The result of the token exchange
        """
        if self._base_client.provider_config.token_endpoint is None:
            raise UnsupportedByProviderError(
                f"The OpenID provider {self._base_client.provider_config.issuer} only supports the implicit flow and does not have a token endpoint"
            )
        if self._base_client.authentication_redirect_uri is None:
            raise ImpossibleOperationError(
                "The client has no redirect_uri configured so the code cannot be exchanged for a token"
            )

        return impl.exchange_code_for_tokens(
            token_endpoint=self._base_client.provider_config.token_endpoint,
            authentication_response=authentication_response,
            redirect_uri=self._base_client.authentication_redirect_uri,
            client_authentication=self._base_client.client_auth,
        )



import logging
import random
import string
from base64 import b64encode
from pathlib import Path
from typing import Mapping

import pytest
import requests
import responses
import yaml
from cryptojwt import JWT, KeyBundle, KeyJar
from cryptojwt.jwk.rsa import new_rsa_key
from responses import matchers

from simple_openid_connect.data import ProviderMetadata

logger = logging.getLogger(__name__)


def rand_str() -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=16))


@pytest.fixture(scope="session")
def jwks() -> KeyBundle:
    """A random JSON-Web-KeySet"""
    key = new_rsa_key()
    bundle = KeyBundle()
    bundle.set([key])
    return bundle


@pytest.fixture(scope="session")
def jwt(jwks) -> JWT:
    """JWT builder based on the generated JWKs"""
    jar = KeyJar()
    jar.add_kb("https://provider.example.com", jwks)
    return JWT(jar, "https://provider.example.com", 3600)


class DummyUserAgent(requests.Session):
    def naviagte_to(self, url: str) -> requests.Response:
        """
        Mimic navigating to the given URL
        """
        return self.get(url, allow_redirects=True)


@pytest.fixture
def user_agent(response_mock) -> DummyUserAgent:
    response_mock.get("https://app.example.com/login-callback")
    response_mock.get("https://app.example.com/logout-callback")
    yield DummyUserAgent()


@pytest.fixture
def response_mock() -> responses.RequestsMock:
    """
    A context manager that mocks and de-mocks http request responses
    """
    with responses.RequestsMock(assert_all_requests_are_fired=False) as mock:
        yield mock


@pytest.fixture
def known_provider_configs(response_mock):
    """
    Mock requests to known OpenID provider config URLs to return static content
    """
    # mafiasi identity
    response_mock.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://identity.mafiasi.de/auth/realms/simple_openid_test","authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth","token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","end_session_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/logout","frontchannel_logout_session_supported":true,"frontchannel_logout_supported":true,"jwks_uri":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs","check_session_iframe":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/login-status-iframe.html","grant_types_supported":["authorization_code","implicit","refresh_token","password","client_credentials","urn:ietf:params:oauth:grant-type:device_code","urn:openid:params:grant-type:ciba"],"response_types_supported":["code","none","id_token","token","id_token token","code id_token","code token","code id_token token"],"subject_types_supported":["public","pairwise"],"id_token_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"id_token_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"id_token_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"userinfo_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512","none"],"request_object_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"request_object_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"response_modes_supported":["query","fragment","form_post","query.jwt","fragment.jwt","form_post.jwt","jwt"],"registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","token_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"token_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"introspection_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"introspection_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"authorization_encryption_alg_values_supported":["RSA-OAEP","RSA-OAEP-256","RSA1_5"],"authorization_encryption_enc_values_supported":["A256GCM","A192GCM","A128GCM","A128CBC-HS256","A192CBC-HS384","A256CBC-HS512"],"claims_supported":["aud","sub","iss","auth_time","name","given_name","family_name","preferred_username","email","acr"],"claim_types_supported":["normal"],"claims_parameter_supported":true,"scopes_supported":["openid","address","web-origins","offline_access","profile","email","roles","phone","microprofile-jwt"],"request_parameter_supported":true,"request_uri_parameter_supported":true,"require_request_uri_registration":true,"code_challenge_methods_supported":["plain","S256"],"tls_client_certificate_bound_access_tokens":true,"revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","revocation_endpoint_auth_methods_supported":["private_key_jwt","client_secret_basic","client_secret_post","tls_client_auth","client_secret_jwt"],"revocation_endpoint_auth_signing_alg_values_supported":["PS384","ES384","RS384","HS256","HS512","ES256","RS256","HS384","ES512","PS256","PS512","RS512"],"backchannel_logout_supported":true,"backchannel_logout_session_supported":true,"device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","backchannel_token_delivery_modes_supported":["poll","ping"],"backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth","backchannel_authentication_request_signing_alg_values_supported":["PS384","ES384","RS384","ES256","RS256","ES512","PS256","PS512","RS512"],"require_pushed_authorization_requests":false,"pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","mtls_endpoint_aliases":{"token_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token","revocation_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/revoke","introspection_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/token/introspect","device_authorization_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/auth/device","registration_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/clients-registrations/openid-connect","userinfo_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/userinfo","pushed_authorization_request_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/par/request","backchannel_authentication_endpoint":"https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/ext/ciba/auth"}}',
    )
    response_mock.get(
        url="https://identity.mafiasi.de/auth/realms/simple_openid_test/protocol/openid-connect/certs",
        content_type="application/json",
        body='{"keys":[{"kid":"JAehgk0O0uzTEht7KGCPVB_urwfsBGe22phHVDZezeo","kty":"RSA","alg":"RS256","use":"sig","n":"n1t4R6mVlBZggtmEM5HohZYg5QjLtjlRVyAMMM93H4WRPpP2Xqj3RrGwP-LZr6lQhk6w8nvPNGnFujWtXb_dBrhZDBh2z0mCNvfjJMp5Ki5sLT9nCbuZ4NkIhfx7qTzQ--GomQ6uKOSuoE12A_r9UopwuCf_1zrx0MIqN_kMeYU2px5yJW-5qbwPL9hcWcfY_Pdz5SFcQ3kDi6GlXA0PN74Ivhs9baVIv60cwB0TytQXGn__GYFZ-K5qAjIymyIy8cPRy9N27Gs29cWF-C56n1pAxwkcPY1SAQgcrspIhntQbZBeGboUx2pnQ_v8LEqVt9ahh52eAbV36YoQqu1DbQ","e":"AQAB","x5c":["MIICszCCAZsCBgGEZpqrijANBgkqhkiG9w0BAQsFADAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwHhcNMjIxMTExMTIxMDIzWhcNMzIxMTExMTIxMjAzWjAdMRswGQYDVQQDDBJzaW1wbGVfb3BlbmlkX3Rlc3QwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCfW3hHqZWUFmCC2YQzkeiFliDlCMu2OVFXIAwwz3cfhZE+k/ZeqPdGsbA/4tmvqVCGTrDye880acW6Na1dv90GuFkMGHbPSYI29+MkynkqLmwtP2cJu5ng2QiF/HupPND74aiZDq4o5K6gTXYD+v1SinC4J//XOvHQwio3+Qx5hTanHnIlb7mpvA8v2FxZx9j893PlIVxDeQOLoaVcDQ83vgi+Gz1tpUi/rRzAHRPK1Bcaf/8ZgVn4rmoCMjKbIjLxw9HL03bsazb1xYX4LnqfWkDHCRw9jVIBCByuykiGe1BtkF4ZuhTHamdD+/wsSpW31qGHnZ4BtXfpihCq7UNtAgMBAAEwDQYJKoZIhvcNAQELBQADggEBAD3aKq2xiLNPlUcmxwtR1FihkEM+NM7VV+m+74Lf9m/lhNzuM00JZAW67gIz2kO5K4XaU84PCRzt3sELma/xVoBs3iW8AhtldTax17g0NZzpMEExzYi0vyihq8xihbx+8XrVwxQ3JBkb/y3b+otDX3xRxBKSCDuxuigL1bX7UzgJ3oe4wD6m8yN7nrMdqSMbspd8lkceKNPziclXnbKPIUwrnxdWMmB41zYcT3y+zHxqk6EDkf3NU+C3gCoRCQoBKcK4X2dwZdUbv0ginU9kROjNXc2+gP53CYDT4fqOEX/jBuljbYb9ziaQ9/5kJZ9SsazgWNjpf4VvhAsH64i7+dE="],"x5t":"PxLqu4M4uNWaUhbfkvI0tGrkhtQ","x5t#S256":"OWVSYSSNFaLyfxV0Lg6XyVDnB2t6sHwtNbzVB2fNCVc"}]}',
    )

    # google
    response_mock.get(
        "https://accounts.google.com/.well-known/openid-configuration",
        content_type="application/json",
        body='{"issuer":"https://accounts.google.com","authorization_endpoint":"https://accounts.google.com/o/oauth2/v2/auth","device_authorization_endpoint":"https://oauth2.googleapis.com/device/code","token_endpoint":"https://oauth2.googleapis.com/token","userinfo_endpoint":"https://openidconnect.googleapis.com/v1/userinfo","revocation_endpoint":"https://oauth2.googleapis.com/revoke","jwks_uri":"https://www.googleapis.com/oauth2/v3/certs","response_types_supported":["code","token","id_token","code token","code id_token","token id_token","code token id_token","none"],"subject_types_supported":["public"],"id_token_signing_alg_values_supported":["RS256"],"scopes_supported":["openid","email","profile"],"token_endpoint_auth_methods_supported":["client_secret_post","client_secret_basic"],"claims_supported":["aud","email","email_verified","exp","family_name","given_name","iat","iss","locale","name","picture","sub"],"code_challenge_methods_supported":["plain","S256"],"grant_types_supported":["authorization_code","refresh_token","urn:ietf:params:oauth:grant-type:device_code","urn:ietf:params:oauth:grant-type:jwt-bearer"]}',
    )


@pytest.fixture
def dummy_provider_config(jwks, response_mock):
    """Mocked responses for the dummy *https://provider.example.com provider*"""
    response_mock.get(
        url="https://provider.example.com/.well-known/openid-configuration",
        json=ProviderMetadata(
            issuer="https://provider.example.com",
            authorization_endpoint="https://provider.example.com/auth",
            token_endpoint="https://provider.example.com/token",
            jwks_uri="https://provider.example.com/jwks",
            userinfo_endpoint="https://provider.example.com/userinfo",
            end_session_endpoint="https://provider.example.com/end-session",
            introspection_endpoint="https://provider.example.com/token-introspection",
            subject_types_supported=["public"],
            id_token_signing_alg_values_supported=["RS256"],
        ).dict(exclude_defaults=True),
    )
    response_mock.get(
        url="https://provider.example.com/jwks",
        body=jwks.jwks(),
        content_type="application/json",
    )


@pytest.fixture
def dummy_auth_response(response_mock):
    """
    Mocked response for the authorization endpoint

    - endpoint url: `https://provider.example.com/auth`
    - client_id: `test-client-id`
    - valid redirect_uri: `https://app.example.com/login-callback`
    - returned code: `code.foobar123`
    """
    response_mock.get(
        url="https://provider.example.com/auth",
        match=[
            matchers.query_param_matcher(
                {
                    "client_id": "test-client-id",
                    "redirect_uri": "https://app.example.com/login-callback",
                    "response_type": "code",
                    "scope": "openid",
                }
            )
        ],
        status=302,
        headers={
            "Location": "https://app.example.com/login-callback?code=code.foobar123"
        },
    )


@pytest.fixture
def dummy_token_response(response_mock):
    """
    Mocked response for the token endpoint

    - endpoint url: `https://provider.example.com/token`
    - client_id: `test-client-id`
    - client_secret: `client-secret`
    - valid redirect_uri: `https://app.example.com/login-callback`
    - expected code: `code.foobar123`

    - returned access_token: `access_token.foobar123`
    - returned id_token: `id_token.user1`
    """
    response_mock.post(
        url="https://provider.example.com/token",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "client_id": "test-client-id",
                    "code": "code.foobar123",
                    "grant_type": "authorization_code",
                    "redirect_uri": "https://app.example.com/login-callback",
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {b64encode(b'test-client-id:test-client-secret').decode()}",
                }
            ),
        ],
        json={
            "access_token": "access_token.foobar123",
            "token_type": "Bearer",
            "id_token": "id_token.user1",
        },
    )


@pytest.fixture
def dumm_userinfo_response(response_mock):
    """
    Mocked response for the userinfo endpoint

    - endpoint url: `https://provider.example.com/userinfo`
    - expected access token: `access_token.foobar123`

    - returned sub: `1`
    - returned username: `user1`
    """
    response_mock.get(
        url="https://provider.example.com/userinfo",
        match=[
            matchers.header_matcher(
                {
                    "Authorization": "Bearer access_token.foobar123",
                }
            )
        ],
        json={"sub": "1", "username": "user1"},
    )


@pytest.fixture
def dummy_end_session_response(response_mock):
    """
    Mocked response for the end-session endpoint

    - endpoint url: `https://provider.example.com/end-session`
    - expected logout redirect uri: `https://app.example.com/logout-callback`

    - redirects to: `https://app.example.com/logout-callback`
    """
    response_mock.get(
        url="https://provider.example.com/end-session",
        match=[
            matchers.query_param_matcher(
                {"post_logout_redirect_uri": "https://app.example.com/logout-callback"}
            )
        ],
        status=302,
        headers={
            "Location": "https://app.example.com/logout-callback",
        },
    )


@pytest.fixture
def dummy_token_introspection_response(response_mock):
    """
    Mocked response for the token introspection endpoint

    - endpoint url: `https://provider.example.com/token-introspection`
    - access token: `access_token.foobar123`
    - client credentials: `test-client-id` & `test-client-secret`

    - returned: `active=true`

    All other access tokens are responded as `active=true`
    """
    response_mock.post(
        url="https://provider.example.com/token-introspection",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "token": "access_token.foobar123",
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {b64encode(b'test-client-id:test-client-secret').decode()}",
                }
            ),
        ],
        json={
            "active": True,
        },
    )
    response_mock.post(
        url="https://provider.example.com/token-introspection",
        match=[
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {b64encode(b'test-client-id:test-client-secret').decode()}",
                }
            )
        ],
        json={
            "active": False,
        },
    )


@pytest.fixture
def secrets() -> Mapping[str, str]:
    path = Path(__file__).parent / "secrets.yml"
    with open(path, "r", encoding="UTF-8") as f:
        return yaml.safe_load(f)



"""
Django AppConfig for this app
"""
import inspect
import logging
from functools import cached_property
from typing import TYPE_CHECKING, List, Optional, Union

from django.apps import AppConfig, apps
from django.conf import settings
from django.core.cache import cache
from django.core.checks import Error, Warning, register
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.shortcuts import resolve_url
from django.utils.module_loading import import_string
from pydantic import BaseModel, ConfigDict

from simple_openid_connect.client import OpenidClient

if TYPE_CHECKING:
    from simple_openid_connect.integrations.django.user_mapping import UserMapper

logger = logging.getLogger(__name__)


class SettingsModel(BaseModel):
    """
    A pydantic model used to validate django settings
    """

    model_config = ConfigDict(from_attributes=True)

    OPENID_ISSUER: str
    OPENID_CLIENT_ID: str
    OPENID_CLIENT_SECRET: Optional[str] = None
    OPENID_SCOPE: str = "openid"
    OPENID_REDIRECT_URI: Optional[str] = "simple_openid_connect:login-callback"
    OPENID_BASE_URI: Optional[str] = None
    OPENID_USER_MAPPER: str = (
        "simple_openid_connect.integrations.django.user_mapping.UserMapper"
    )


class OpenidAppConfig(AppConfig):
    name = "simple_openid_connect.integrations.django"
    label = "simple_openid_connect_django"

    default_auto_field = "django.db.models.BigAutoField"

    def ready(self) -> None:
        """
        Called when django starts.

        Performs settings validation and raises ImproperlyConfigured if necessary.
        """
        super().ready()

        # assert that are settings are as required
        try:
            _ = self.safe_settings
            _ = self.user_mapper
        except Exception as e:
            raise ImproperlyConfigured(
                f"django settings are invalid for openid usage: {e}"
            ) from e

    @classmethod
    def get_instance(cls) -> "OpenidAppConfig":
        """
        Retrieve the currently used instance from django's app registry
        """
        instance = apps.get_app_config(cls.label)
        assert isinstance(instance, OpenidAppConfig)
        return instance

    @cached_property
    def safe_settings(self) -> SettingsModel:
        """
        type-validated version of django settings
        """
        return SettingsModel.from_orm(settings)

    @cached_property
    def user_mapper(self) -> "UserMapper":
        """
        A UserMapper instance of which the actual implementation is configured via django settings.
        """
        from simple_openid_connect.integrations.django.user_mapping import UserMapper

        cls = import_string(self.safe_settings.OPENID_USER_MAPPER)
        if not issubclass(cls, UserMapper):
            raise ImproperlyConfigured(
                "OPENID_USER_MAPPER setting does not point to a subclass of UserMapper",
                cls,
            )
        return cls()  # type: ignore

    def get_client(
        self, own_base_uri: Union[HttpRequest, str, None] = None
    ) -> OpenidClient:
        """
        Get an `OpenidClient` instance that is appropriate for usage in django.

        It is automatically configured via django settings.

        :param own_base_uri: The base url of this application which will be used to construct a redirect_uri back to it.
            Can also be the current request in which case `{scheme}://{host}` of it will be used as the base url.
            If this parameter is not given, only the `OPENID_BASE_URI` setting is used.
            In any case, if the `OPENID_BASE_URI` setting is set, it will be used instead.

        :raises ImproperlyConfigured: when no *own_base_uri* is given and the `OPENID_BASE_URI` is also None
        """
        # use a cached client instance if one exists or create a new one if not
        client = cache.get("openid_client")  # type: OpenidClient
        if client is None:
            # determine base_uri of this app
            if self.safe_settings.OPENID_REDIRECT_URI is not None:
                if self.safe_settings.OPENID_BASE_URI is not None:
                    own_base_uri = self.safe_settings.OPENID_BASE_URI
                else:
                    if own_base_uri is None:
                        raise ImproperlyConfigured(
                            "either a value for own_base_uri must be given or the django setting OPENID_BASE_URI must be filled"
                        )
                    elif isinstance(own_base_uri, HttpRequest):
                        own_base_uri = (
                            f"{own_base_uri.scheme}://{own_base_uri.get_host()}"
                        )

                relative_redirect_uri = resolve_url(
                    self.safe_settings.OPENID_REDIRECT_URI
                )
                redirect_uri = f"{own_base_uri}{relative_redirect_uri}"
            else:
                redirect_uri = None

            # create a new client instance and cache it
            client = OpenidClient.from_issuer_url(
                url=self.safe_settings.OPENID_ISSUER,
                authentication_redirect_uri=redirect_uri,
                client_id=self.safe_settings.OPENID_CLIENT_ID,
                client_secret=self.safe_settings.OPENID_CLIENT_SECRET,
                scope=self.safe_settings.OPENID_SCOPE,
            )
            cache.set("openid_client", client)

        return client


@register  # type: ignore
def check_middleware(*args, **kwargs) -> List[Union[Warning, Error]]:
    from simple_openid_connect.integrations.django.middleware import (
        TokenVerificationMiddleware,
    )

    if (
        hasattr(settings, "OPENID_REDIRECT_URI")
        and settings.OPENID_REDIRECT_URI is None  # type: ignore
    ):
        # redirect uri is set to None,
        # so the library is probably used as a resource server,
        # therefore we don't need the middleware
        return []

    has_token_verification_middleware = False
    for middleware in settings.MIDDLEWARE:
        try:
            middleware_cls = import_string(middleware)
        except ImportError:
            continue
        if inspect.isclass(middleware_cls) and issubclass(
            middleware_cls, TokenVerificationMiddleware
        ):
            has_token_verification_middleware = True
            break

    if not has_token_verification_middleware:
        return [
            Warning(
                "Access tokens are not checked for expiry and users are not automatically logged out when their access token expires.",
                hint="Add 'simple_openid_connect.integrations.django.middleware.TokenVerificationMiddleware' to MIDDLEWARE.",
                obj="simple_openid_connect",
                id="simple_openid_connect.W001",
            )
        ]
    else:
        return []



from typing import Any

import pytest
import requests
from django.contrib.auth.models import User
from django.http import (
    HttpResponse,
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
)
from django.test.client import Client as DjangoClient
from furl import furl

from simple_openid_connect.integrations.django import models


@pytest.fixture
def dummy_provider_settings(settings):
    settings.ALLOWED_HOSTS.append("app.example.com")
    settings.OPENID_SCOPE = "openid"
    settings.OPENID_ISSUER = "https://provider.example.com"
    settings.OPENID_BASE_URI = "https://app.example.com"
    settings.OPENID_CLIENT_ID = "client-id"
    settings.OPENID_CLIENT_SECRET = "client-secret"


class DynClient(DjangoClient):
    def request(self, **request: Any) -> HttpResponse:
        # handle internal hosts normally
        from django.conf import settings

        if (
            "SERVER_NAME" not in request.keys()
            or request["SERVER_NAME"] in settings.ALLOWED_HOSTS
        ):
            return super().request(**request)

        # handle external hosts via request
        url = furl(
            scheme=request["wsgi.url_scheme"],
            host=request["SERVER_NAME"],
            port=request["SERVER_PORT"] if request["SERVER_PORT"] != "80" else None,
            path=request["PATH_INFO"],
            query=request["QUERY_STRING"],
        )
        response = requests.request(
            method=request["REQUEST_METHOD"],
            url=str(url),
            allow_redirects=False,
        )  # type: requests.Response
        base_kwargs = {
            "content": response.content,
            "status": response.status_code,
            "reason": response.reason,
            "charset": response.apparent_encoding,
            "headers": response.headers,
        }
        if response.status_code == 302:
            return HttpResponseRedirect(
                redirect_to=response.headers["Location"], **base_kwargs
            )
        elif response.status_code == 301:
            return HttpResponsePermanentRedirect(
                redirect_to=response.headers["Location"], **base_kwargs
            )
        return HttpResponse(**base_kwargs)


@pytest.fixture
def dyn_client(client) -> DynClient:
    """A client that automatically distinguishes between internal django calls and calls to external domains"""
    return DynClient()


@pytest.fixture
def test_user(db) -> User:
    user = User.objects.create_user("user1")
    openid_user = models.OpenidUser.objects.create(sub="1", user=user)
    return user



"""
Django settings for django_test_project project.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-foobar123"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

INTERNAL_IPS = [
    "127.0.0.1",
    "::1",
]

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "debug_toolbar",
    "simple_openid_connect.integrations.django",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "django_test_project.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "django_test_project.wsgi.application"

# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

LOGIN_REDIRECT_URL = "default-after-login"

OPENID_BASE_URI = "https://app.example.com"
OPENID_ISSUER = "https://provider.example.com"
OPENID_CLIENT_ID = "test-client-id"
OPENID_CLIENT_SECRET = "test-client-secret"
OPENID_SCOPE = "openid profile email"

LOGIN_URL = "simple_openid_connect:login"



import logging
from datetime import datetime, timezone
from typing import Callable

from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse

from simple_openid_connect.data import TokenSuccessResponse
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.integrations.django.models import OpenidSession

logger = logging.getLogger(__name__)


class TokenVerificationMiddleware:
    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        response = self.get_response(request)
        openid_session_id = request.session.get("openid_session")
        if not openid_session_id:
            return response

        openid_session = OpenidSession.objects.get(id=openid_session_id)
        refresh_token = openid_session.refresh_token
        session_valid_until = openid_session.access_token_expiry
        access_token_valid = (
            session_valid_until is not None
            and session_valid_until > datetime.now(timezone.utc)
        )
        if access_token_valid:
            return response

        logger.debug("access token expired, trying to refresh")
        client = OpenidAppConfig.get_instance().get_client(request)
        exchange_response = client.exchange_refresh_token(refresh_token)
        if isinstance(exchange_response, TokenSuccessResponse):
            openid_session.update_session(exchange_response)
            openid_session.save()
            return response
        else:
            return HttpResponseRedirect(resolve_url(settings.LOGIN_URL))



"""
DRF Authentication classes

See the `DRF documentation on Setting the authentication scheme <https://www.django-rest-framework.org/api-guide/authentication/#setting-the-authentication-scheme>`_ about how to use the classes contained here.
"""

import logging
from typing import Any, Optional, Tuple, Union

from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed

from simple_openid_connect.data import (
    JwtAccessToken,
    TokenIntrospectionErrorResponse,
    TokenIntrospectionSuccessResponse,
)
from simple_openid_connect.exceptions import ValidationError
from simple_openid_connect.integrations.django import models
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.integrations.django.user_mapping import FederatedUserData

logger = logging.getLogger(__name__)


class AuthenticatedViaToken:
    """
    A marker that is attached as ``request.auth`` on successful authentication and which holds well formatted
    information about that authentication.
    """

    def __init__(
        self,
        token: str,
        user_data: FederatedUserData,
    ):
        self.token = token
        self.user_data = user_data

    def __str__(self) -> str:
        return self.token


class AccessTokenAuthentication(BaseAuthentication):
    """
    An authentication scheme that interprets ``Authorization: Bearer ...`` http headers as access tokens.
    """

    def authenticate(
        self, request: HttpRequest
    ) -> Union[Tuple[Any, AuthenticatedViaToken], None]:
        # abort if no authentication is intended
        if "Authorization" not in request.headers.keys() or not request.headers[
            "Authorization"
        ].startswith("Bearer "):
            return None

        oidc_client = OpenidAppConfig.get_instance().get_client(request)
        raw_token = request.headers["Authorization"].split(" ", 1)[1]

        # handle access token while not verifying scopes because those are verified by a permission class
        try:
            (
                user,
                userinfo,
            ) = OpenidAppConfig.get_instance().user_mapper.handle_federated_access_token(
                raw_token, oidc_client, required_scopes=""
            )
            return user, AuthenticatedViaToken(raw_token, userinfo)
        except ValidationError:
            raise AuthenticationFailed()

    def authenticate_header(self, request: HttpRequest) -> str:
        return "Bearer"



"""
Default implementations for mapping tokens to user objects.

This implementation can be overriden by extending the :class:`UserMapper` class and then setting the django settings
variable ``OPENID_USER_MAPPER`` to an import string pointing to the newly created class.
"""
import logging
from typing import Any, Tuple, Union

from django.contrib.auth.models import AbstractBaseUser, AbstractUser
from django.db import transaction

from simple_openid_connect.client import OpenidClient
from simple_openid_connect.data import (
    IdToken,
    JwtAccessToken,
    TokenIntrospectionErrorResponse,
    TokenIntrospectionSuccessResponse,
    UserinfoSuccessResponse,
)
from simple_openid_connect.exceptions import ValidationError
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.integrations.django.models import OpenidUser

logger = logging.getLogger(__name__)

FederatedUserData = Union[
    IdToken, UserinfoSuccessResponse, TokenIntrospectionSuccessResponse, JwtAccessToken
]
"Type alias for the different classes which can provide information about a federated user."


class UserMapper:
    """
    A base class which is responsible for mapping federated users into the local system.

    This class is used in different parts of *simple_openid_connect* and the callgraph is sketched below:

    .. code-block:: text

       ┌──────────────────────────────────┐   ┌──────────────────────────────────┐   ┌───────────────────┐
       │ @access_token_required decorator │   │ DRF AccessTokenAuthentication    │   │ LoginCallbackView │
       └────────────────┬─────────────────┘   └─────────────────┬────────────────┘   └─────────┬─────────┘
                        │                                       │                              │
                        └───────────────────┬───────────────────┘                              │
                                            ↓                                                  │
                     ┌──────────────────────┴─────────────────────┐                            │
                     │ UserMapper.handle_federated_access_token() │                            │
                     └──────────────────────┬─────────────────────┘                            │
                                            │                                                  │
                                            └──────────────────────────┬───────────────────────┘
                                                                       ↓
                                                  ┌────────────────────┴───────────────────┐
                                                  │ UserMapper.handle_federated_userinfo() │
                                                  └────────────────────┬───────────────────┘
                                                                       ↓
                                                      ┌────────────────┴────────────────┐
                                                      │ UserMapper.automap_user_attrs() │
                                                      └─────────────────────────────────┘
    """

    def handle_federated_userinfo(self, user_data: FederatedUserData) -> Any:
        """
        Entry point for dynamically creating or updating user data based on information obtained through OpenID.

        The function automatically creates a new user model instance if the user is unknown or updates the locally
        stored user information based on the federated data.

        :param user_data: Information about the user.

        :returns: An instance of the applications user model.
        """
        # validate that user data contains at least a user id
        if user_data.sub is None:
            raise ValidationError(
                "could not map user to token because the issuer did not return a 'sub' claim in its token introspection response"
            )

        with transaction.atomic():
            openid_user = OpenidUser.objects.get_or_create_for_sub(user_data.sub)
            user = openid_user.user
            self.automap_user_attrs(user, user_data)
            user.save()
            return user

    def handle_federated_access_token(
        self,
        access_token: str,
        oidc_client: OpenidClient,
        required_scopes: Union[str, None] = None,
    ) -> Tuple[Any, FederatedUserData]:
        """
        Entry point for dynamically creating or updating user data based on an access token which was provided by a user.

        This method inspects the token and then calls into :meth:`UserMapper.handle_federated_userinfo()` once more information
        about the user is available.

        :param access_token: The raw access token that was passed to this application which should identify the user.
        :param oidc_client: An OpenID client which is used to access the OpenID providers signing keys or to introspect the token if necessary.
        :param required_scopes: Scopes to which the access token is required to have access.
            If ``None`` is passed, the default scopes from django settings ``OPENID_SCOPE`` are used.
            Pass an empty string if no scopes are required.

        :returns: An instance of the applications user model as well as additional data about the user.

        :raises ValidationError: If the passed token cannot be validated or is decidedly invalid.
        """
        if required_scopes is None:
            required_scopes = OpenidAppConfig.get_instance().safe_settings.OPENID_SCOPE

        # try to parse the raw token as JWT
        user_data = (
            None
        )  # type: JwtAccessToken | TokenIntrospectionSuccessResponse | None
        try:
            # parse an validate the general token structure
            token = JwtAccessToken.parse_jwt(
                access_token,
                oidc_client.provider_keys,
            )
            token.validate_extern(oidc_client.provider_config.issuer)

            # validate token scope for required access
            if required_scopes != "":
                if token.scope is None:
                    raise ValidationError(
                        "token does not contain required scopes claim"
                    )
                elif any(
                    i_scope not in token.scope.split(" ")
                    for i_scope in required_scopes.split(" ")
                ):
                    raise ValidationError(
                        f"token has access to scopes '{token.scope}' but '{required_scopes}' are required"
                    )

            # the token is determined to be valid, so we can use it as user_data
            user_data = token

        # fall back to introspecting the token at the issuer
        except Exception:
            logger.debug(
                "could not parse access token as JWT, falling back to calling the providers token introspection endpoint"
            )
            introspect_response = oidc_client.introspect_token(access_token)
            if isinstance(introspect_response, TokenIntrospectionErrorResponse):
                logger.critical(
                    "could not introspect token for validity: %s", introspect_response
                )
                raise ValidationError(
                    f"could not introspect token at the issuer: {introspect_response}"
                )

            # fail if the token is expired
            if not introspect_response.active:
                raise ValidationError("token is expired")

            # validate token scope for required access
            if introspect_response.scope is None:
                logger.error(
                    "could not determine access token access because the issuer did not return the tokens scope during token instrospection"
                )
                raise ValidationError(
                    "could not determine token scope because the issuer did not return the tokens scope during token introspection"
                )
            elif any(
                i_scope not in introspect_response.scope.split(" ")
                for i_scope in required_scopes.split(" ")
            ):
                raise ValidationError(
                    f"token has access to scopes '{introspect_response.scope}' but '{required_scopes}' are required"
                )

            # the token is determined to be valid, so we can use it as user_data
            user_data = introspect_response

        return self.handle_federated_userinfo(user_data), user_data

    def automap_user_attrs(
        self,
        user: "AbstractBaseUser",
        user_data: FederatedUserData,
    ) -> None:
        """
        Inspect the given user instance model, discover its attributes based on some heuristics and set their values
        from the passed user information.

        .. note::

           ``user.save()`` is not automatically called by this method to allow extending it via class inheritance
            without causing multiple database operations.

        :param user: The user instance on which attributes should be set
        :param user_data: Information about the user which was made available through OpenID.
        """
        if isinstance(user, AbstractUser):
            # username
            if hasattr(user_data, "preferred_username"):
                setattr(user, user.USERNAME_FIELD, user_data.preferred_username)
            elif hasattr(user_data, "username"):
                setattr(user, user.USERNAME_FIELD, user_data.username)
            elif hasattr(user_data, "sub"):
                setattr(user, user.USERNAME_FIELD, user_data.sub)
            else:
                logger.warning(
                    "Could not determine a username from federated user data. Creating more than one user will probably fail because the users username attribute is mapped to be empty and django enforces a unique-constraint on usernames."
                )
            # email
            if hasattr(user_data, "email"):
                setattr(user, user.EMAIL_FIELD, user_data.email)
            # given name
            if hasattr(user_data, "given_name") and hasattr(user, "first_name"):
                user.first_name = user_data.given_name
            # family name
            if hasattr(user_data, "family_name") and hasattr(user, "last_name"):
                user.last_name = user_data.family_name



import json

from django.shortcuts import resolve_url


def test_unauth_access(client):
    response = client.get(resolve_url("access-token-protected-view"))
    assert response.status_code == 401
    assert response.headers["WWW-Authenticate"] == "Bearer"


def test_invalid_token_text(
    client,
    dummy_provider_settings,
    dummy_provider_config,
    dummy_token_introspection_response,
    response_mock,
):
    # act
    response = client.get(
        resolve_url("access-token-protected-view"),
        HTTP_Authorization="Bearer invalid-token",
    )

    # assert
    assert response.status_code == 401
    assert response.headers["WWW-Authenticate"] == "Bearer"
    assert (
        response.content
        == b"the used access token is not valid or does not grant enough access"
    )


def test_invalid_token_json(
    client,
    dummy_provider_settings,
    dummy_provider_config,
    dummy_token_introspection_response,
):
    # act
    response = client.get(
        resolve_url("access-token-protected-view"),
        HTTP_Authorization="Bearer access_token.invalid",
        HTTP_Accept="application/json",
    )

    # assert
    assert response.status_code == 401
    assert response.headers["WWW-Authenticate"] == "Bearer"
    assert response.headers["Content-Type"] == "application/json"
    json_response = json.loads(response.content)
    assert json_response["error"] == "invalid_token"
    assert (
        json_response["error_description"]
        == "the used access token is not valid or does not grant enough access"
    )



from typing import TYPE_CHECKING, Union

from simple_openid_connect.data import TokenErrorResponse, TokenSuccessResponse
from simple_openid_connect.exceptions import UnsupportedByProviderError
from simple_openid_connect.flows import direct_access_grant as impl

if TYPE_CHECKING:
    from simple_openid_connect.client import OpenidClient


class DirectAccessGrantClient:
    """
    A client that implements *Direct Access Grant* (or *Resource Owner Password Credentials Grant*) functionality.

    It is exposed via :data:`OpenidClient.authorization_code_flow <simple_openid_connect.client.OpenidClient.direct_access_grant>`.
    """

    def __init__(self, base_client: "OpenidClient"):
        self._base_client = base_client

    def authenticate(
        self, username: str, password: str
    ) -> Union[TokenSuccessResponse, TokenErrorResponse]:
        """
        Exchange a given username and password for access, refresh and id tokens.

        :returns: The result of the exchange
        """
        if self._base_client.provider_config.token_endpoint is None:
            raise UnsupportedByProviderError(
                f"The OpenID provider {self._base_client.provider_config.issuer} only seems to support the implicit flow and does not have a token endpoint"
            )

        return impl.authenticate(
            token_endpoint=self._base_client.provider_config.token_endpoint,
            scope=self._base_client.scope,
            username=username,
            password=password,
            client_authentication=self._base_client.client_auth,
        )



from base64 import b64encode

import pytest
from responses import matchers

from simple_openid_connect.client_authentication import ClientSecretBasicAuth
from simple_openid_connect.flows import client_credentials_grant


@pytest.fixture
def dummy_token_response(response_mock):
    response_mock.post(
        url="https://provider.example.com/token",
        match=[
            matchers.urlencoded_params_matcher(
                {
                    "grant_type": "client_credentials",
                    "scope": "openid",
                }
            ),
            matchers.header_matcher(
                {
                    "Authorization": f"Basic {b64encode(b'client-id:client-secret').decode()}",
                }
            ),
        ],
        json={
            "access_token": "access_token.foobar123",
            "token_type": "Bearer",
            "id_token": "id_token.foobar123",
        },
    )


def test_auth_exchange(user_agent, dummy_token_response):
    # act
    response = client_credentials_grant.authenticate(
        "https://provider.example.com/token",
        "openid",
        ClientSecretBasicAuth("client-id", "client-secret"),
    )

    # assert
    assert response.access_token



# Generated by Django 4.2.5 on 2023-11-13 08:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("simple_openid_connect_django", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="openiduser",
            name="_id_token",
        ),
        migrations.AddField(
            model_name="openidsession",
            name="_id_token",
            field=models.TextField(
                default="{}",
                verbose_name="json representation of this sessions is token",
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 3.2.16 on 2022-11-16 10:48

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="OpenidUser",
            fields=[
                (
                    "sub",
                    models.CharField(
                        help_text="subject identifier of this user (userid)",
                        max_length=255,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "_id_token",
                    models.TextField(
                        help_text="json representation of the most current id token"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="openid",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="OpenidSession",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sid",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="session id",
                        max_length=256,
                    ),
                ),
                ("scope", models.CharField(max_length=256)),
                ("access_token", models.TextField()),
                ("access_token_expiry", models.DateTimeField(null=True)),
                ("refresh_token", models.TextField(blank=True)),
                ("refresh_token_expiry", models.DateTimeField(null=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sessions",
                        to="simple_openid_connect_django.openiduser",
                    ),
                ),
            ],
        ),
    ]



"""
View functions which handle openid authentication and their related callbacks
"""
import logging
from http import HTTPStatus
from typing import Mapping

from django.conf import settings
from django.contrib.auth import login, logout
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.shortcuts import resolve_url
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.cache import cache_control

from simple_openid_connect.data import (
    FrontChannelLogoutNotification,
    IdToken,
    RpInitiatedLogoutRequest,
    TokenSuccessResponse,
)
from simple_openid_connect.integrations.django.apps import OpenidAppConfig
from simple_openid_connect.integrations.django.models import OpenidSession

logger = logging.getLogger(__name__)


class InitLoginView(View):
    """
    The view which handles initiating a login.

    It essentially redirects the user agent to the Openid provider.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        logout(request)
        if "next" in request.GET.keys():
            request.session["login_redirect_url"] = request.GET["next"]
        client = OpenidAppConfig.get_instance().get_client(request)
        redirect = client.authorization_code_flow.start_authentication()
        return HttpResponseRedirect(redirect)


class LoginCallbackView(View):
    """
    The view which handles login callbacks.

    It handles an authentication response from the Openid provider that is encoded in the current url by either logging
    the user in or rendering the error.

    Error rendering can be customized by overwriting the template *simple_openid_connect/login_failed.html* which
    receives the context `token_response` of type :class:`TokenErrorResponse <simple_openid_connect.data.TokenErrorResponse>`.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        client = OpenidAppConfig.get_instance().get_client(request)

        # exchange the passed code for tokens
        token_response = client.authorization_code_flow.handle_authentication_result(
            current_url=request.get_full_path(),
        )
        if not isinstance(token_response, TokenSuccessResponse):
            return TemplateResponse(
                request,
                "simple_openid_connect/login_failed.html",
                {
                    "token_response": token_response,
                },
                status=HTTPStatus.UNAUTHORIZED,
            )

        # validate the received tokens
        id_token = IdToken.parse_jwt(token_response.id_token, client.provider_keys)
        id_token.validate_extern(
            client.provider_config.issuer, client.client_auth.client_id
        )

        # handle federated user information (create a new user if necessary or update local info) and log the user in
        user = OpenidAppConfig.get_instance().user_mapper.handle_federated_userinfo(
            id_token
        )
        openid_session = user.openid.update_session(token_response, id_token)
        request.session["openid_session"] = openid_session.id
        login(request, user, backend=settings.AUTHENTICATION_BACKENDS[0])

        # redirect to the next get parameter if present, otherwise to the configured default
        if "login_redirect_url" in request.session.keys():
            return HttpResponseRedirect(
                redirect_to=request.session["login_redirect_url"]
            )
        else:
            return HttpResponseRedirect(
                redirect_to=resolve_url(settings.LOGIN_REDIRECT_URL)
            )


class LogoutView(View):
    """
    The view which handles logging a user out.
    """

    def get(self, request: HttpRequest) -> HttpResponse:
        session_id = request.session.get("openid_session")
        logout(request)
        client = OpenidAppConfig.get_instance().get_client(request)

        if settings.LOGOUT_REDIRECT_URL is not None:
            openid_session = (
                OpenidSession.objects.get(id=session_id) if session_id else None
            )

            logout_request = RpInitiatedLogoutRequest(
                post_logout_redirect_uri=request.build_absolute_uri(
                    resolve_url(settings.LOGOUT_REDIRECT_URL)
                )
            )
            if openid_session is not None and openid_session.raw_id_token is not None:
                logout_request.id_token_hint = openid_session.raw_id_token
            else:
                logout_request.client_id = client.client_auth.client_id
        else:
            logout_request = None

        return HttpResponseRedirect(client.initiate_logout(logout_request))


class FrontChannelLogoutNotificationView(View):
    """
    A view which handles Openid front-channel logout notifications by logging out the current session
    """

    @method_decorator(cache_control(no_store=True))
    def get(self, request: HttpRequest) -> HttpResponse:
        logout(request)
        return HttpResponse(status=200)



"""
simple_openid_connect URL Configuration

These urls should be included into your project wherever you like via::

    urlpatterns = [
        ...
        path("auth/openid/", include("simple_openid_connect.integrations.django.urls")),
    ]

"""
from django.urls import path

from . import views

app_name = "simple_openid_connect"
urlpatterns = [
    path("login/", views.InitLoginView.as_view(), name="login"),
    path("login-callback/", views.LoginCallbackView.as_view(), name="login-callback"),
    path("logout/", views.LogoutView.as_view(), name="logout"),
    path(
        "logout/frontchannel-notify/",
        views.FrontChannelLogoutNotificationView.as_view(),
        name="logout-frontchannel-notify",
    ),
]
